import hashlib
import json
import os
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import FAST_BAKERY_URL, get_pinned_conda_libs
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.plugins.aws.batch.batch_decorator import BatchDecorator
from metaflow.plugins.kubernetes.kubernetes_decorator import KubernetesDecorator
from metaflow.plugins.pypi.conda_decorator import CondaStepDecorator
from metaflow.plugins.pypi.conda_environment import CondaEnvironment
from metaflow.plugins.pypi.pypi_decorator import PyPIStepDecorator
from metaflow import decorators

from .fast_bakery import FastBakery, FastBakeryApiResponse, FastBakeryException

BAKERY_METAFILE = ".imagebakery-cache"

import fcntl
import json
import os
from concurrent.futures import ThreadPoolExecutor
from functools import wraps


def cache_request(cache_file):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            call_args = kwargs.copy()
            call_args.update(zip(func.__code__.co_varnames, args))
            call_args.pop("self", None)
            call_args.pop("ref", None)
            # invalidate cache when moving from one deployment to another
            call_args.update({"fast_bakery_url": FAST_BAKERY_URL})
            cache_key = hashlib.md5(
                json.dumps(call_args, sort_keys=True).encode("utf-8")
            ).hexdigest()

            try:
                with open(cache_file, "r") as f:
                    cache = json.load(f)
                    if cache_key in cache:
                        return FastBakeryApiResponse(cache[cache_key])
            except (FileNotFoundError, json.JSONDecodeError):
                cache = {}

            result = func(*args, **kwargs)

            try:
                with open(cache_file, "r+") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    try:
                        f.seek(0)
                        cache = json.load(f)
                    except json.JSONDecodeError:
                        cache = {}

                    cache[cache_key] = result.response

                    f.seek(0)
                    f.truncate()
                    json.dump(cache, f)
            except FileNotFoundError:
                # path to cachefile might not exist.
                os.makedirs(os.path.dirname(cache_file), exist_ok=True)
                with open(cache_file, "w") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    json.dump({cache_key: result.response}, f)

            return result

        return wrapper

    return decorator


class DockerEnvironmentException(MetaflowException):
    headline = "Ran into an error while baking image"

    def __init__(self, msg):
        super(DockerEnvironmentException, self).__init__(msg)


class DockerEnvironment(MetaflowEnvironment):
    TYPE = "fast-bakery"
    _filecache = None

    def __init__(self, flow):
        self.skipped_steps = set()
        self.flow = flow

        self.results = {}
        self.images_baked = 0

    def set_local_root(self, local_root):
        self.local_root = local_root

    def decospecs(self):
        # Due to conflicts with the CondaEnvironment fallback and bakery,
        # we can not simply attach 'conda' or 'pypi' to all steps here.
        # Instead we do this on a per-step basis in init_environment
        return ("fast_bakery_internal",) + super().decospecs()

    def validate_environment(self, logger, datastore_type):
        self.datastore_type = datastore_type
        self.logger = logger

        # Avoiding circular imports.
        from metaflow.plugins import DATASTORES

        self.datastore = [d for d in DATASTORES if d.TYPE == self.datastore_type][0]

        # Mixing @pypi/@conda in a single step is not supported yet
        for step in self.flow:
            if sum(1 for deco in step.decorators if _is_env_deco(deco)) > 1:
                raise MetaflowException(
                    "Mixing and matching PyPI packages and Conda packages within a\n"
                    "step is not yet supported. Use one of @pypi or @conda only for the *%s* step."
                    % step.name
                )

    def init_environment(self, echo):
        self.skipped_steps = {
            step.name for step in self.flow if not _step_executes_remotely(step)
        }
        # Attach environment decorator as needed. This is done on a step-by-step basis
        # as we require a conda decorator for fallback steps, but prefer pypi for the baked ones.
        for step in self.flow:
            if not _step_has_environment_deco(step):
                if step.name in self.skipped_steps:
                    # Conda fallback requires a conda decorator as the default for a step
                    decorators._attach_decorators_to_step(step, ["conda"])
                else:
                    # We default to PyPI for steps that are going to be baked.
                    decorators._attach_decorators_to_step(step, ["pypi"])
                # Initialize the decorator we attached.
                # This is crucial for the conda decorator to work properly in the fallback environment
                for deco in step.decorators:
                    if _is_env_deco(deco):
                        deco.step_init(
                            self.flow,
                            None,  # not passing graph as it is not available, and not required by conda/pypi decorators
                            step.name,
                            step.decorators,
                            self,
                            self.datastore,
                            echo,
                        )

        steps_to_bake = [
            step for step in self.flow if step.name not in self.skipped_steps
        ]
        if steps_to_bake:
            self.logger("🚀 Baking container image(s) ...")
            start_time = time.time()
            self.results = self._bake(steps_to_bake)
            for step in self.flow:
                for d in step.decorators:
                    if _is_remote_deco(d):
                        d.attributes["image"] = self.results[step.name].container_image
                        d.attributes["executable"] = self.results[step.name].python_path
            if self.images_baked > 0:
                bake_time = time.time() - start_time
                self.logger(
                    f"🎉 All container image(s) baked in {bake_time:.2f} seconds!"
                )
            else:
                self.logger("🎉 All container image(s) baked!")

        if self.skipped_steps:
            self.delegate = CondaEnvironment(self.flow)
            self.delegate.set_local_root(self.local_root)
            self.delegate.validate_environment(echo, self.datastore_type)
            self.delegate.init_environment(echo, self.skipped_steps)

    def _bake(self, steps) -> Dict[str, FastBakeryApiResponse]:
        metafile_path = get_fastbakery_metafile_path(self.local_root, self.flow.name)
        logger_lock = threading.Lock()

        @cache_request(metafile_path)
        def _cached_bake(
            ref=None,
            python=None,
            pypi_packages=None,
            conda_packages=None,
            base_image=None,
        ):
            try:
                bakery = FastBakery(url=FAST_BAKERY_URL)
                bakery._reset_payload()
                bakery.python_version(python)
                bakery.pypi_packages(pypi_packages)
                bakery.conda_packages(conda_packages)
                bakery.base_image(base_image)
                # bakery.ignore_cache()

                with logger_lock:
                    self.logger(f"🍳 Baking [{ref}] ...")
                    self.logger(f"     🐍 Python: {python}")

                    if pypi_packages:
                        self.logger(f"     📦 PyPI packages:")
                        for package, version in pypi_packages.items():
                            self.logger(f"        🔧 {package}: {version}")

                    if conda_packages:
                        self.logger(f"     📦 Conda packages:")
                        for package, version in conda_packages.items():
                            self.logger(f"        🔧 {package}: {version}")

                    self.logger(f"     🏗️  Base image: {base_image}")

                start_time = time.time()
                res = bakery.bake()
                # TODO: Get actual bake time from bakery
                bake_time = time.time() - start_time

                with logger_lock:
                    self.logger(f"🏁 Baked [{ref}] in {bake_time:.2f} seconds!")
                self.images_baked += 1
                return res
            except FastBakeryException as ex:
                raise DockerEnvironmentException(f"Bake [{ref}] failed: {str(ex)}")

        def prepare_step(step):
            base_image = next(
                (
                    d.attributes.get("image")
                    for d in step.decorators
                    if isinstance(d, (KubernetesDecorator))
                ),
                None,
            )
            dependencies = next(
                (d for d in step.decorators if _is_env_deco(d)),
                None,
            )
            python = next(
                (
                    d.attributes["python"]
                    for d in step.decorators
                    if isinstance(d, CondaStepDecorator)
                ),
                None,
            )
            pypi_deco = next(
                (d for d in step.decorators if isinstance(d, PyPIStepDecorator)), None
            )
            # if pypi decorator is set and user has specified a python version, we must create a new environment.
            # otherwise rely on the base environment
            if pypi_deco is not None:
                python = (
                    pypi_deco.attributes["python"]
                    if pypi_deco.is_attribute_user_defined("python")
                    else None
                )

            packages = get_pinned_conda_libs(python, self.datastore_type)
            packages.update(dependencies.attributes["packages"] if dependencies else {})

            return {
                "python": python,
                "pypi_packages": (
                    packages if isinstance(dependencies, PyPIStepDecorator) else None
                ),
                "conda_packages": (
                    packages if isinstance(dependencies, CondaStepDecorator) else None
                ),
                "base_image": base_image,
            }

        with ThreadPoolExecutor() as executor:
            prepared_args = list(executor.map(prepare_step, steps))
            for i, args in enumerate(prepared_args, 1):
                args["ref"] = f"#{i:02d}"
            futures = [executor.submit(_cached_bake, **args) for args in prepared_args]
            results = {}
            for step, future in zip(steps, futures):
                results[step.name] = future.result()

            return results

    def executable(self, step_name, default=None):
        if step_name in self.skipped_steps:
            return self.delegate.executable(step_name, default)
        # default is set to the right executable
        if default is not None:
            return default
        if default is None and step_name in self.results:
            # try to read pythonpath from results. This can happen immediately after baking.
            return self.results[step_name].python_path
        # we lack a default and baking results. fallback to parent executable.
        return super().executable(step_name, default)

    def interpreter(self, step_name):
        if step_name in self.skipped_steps:
            return self.delegate.interpreter(step_name)
        return None

    def is_disabled(self, step):
        for decorator in step.decorators:
            # @conda decorator is guaranteed to exist thanks to self.decospecs
            if decorator.name in ["conda", "pypi"]:
                # handle @conda/@pypi(disabled=True)
                disabled = decorator.attributes["disabled"]
                return str(disabled).lower() == "true"
        return False

    def pylint_config(self):
        config = super().pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    def bootstrap_commands(self, step_name, datastore_type):
        if step_name in self.skipped_steps:
            return self.delegate.bootstrap_commands(step_name, datastore_type)
        # Bootstrap conda and execution environment for step
        # we set the environment flag for skipping bootstrap dependencies, as these are
        # provided in all baked images.
        return [
            "export METAFLOW_SKIP_INSTALL_DEPENDENCIES=$FASTBAKERY_IMAGE",
        ] + super().bootstrap_commands(step_name, datastore_type)


def get_fastbakery_metafile_path(local_root, flow_name):
    return os.path.join(local_root, flow_name, BAKERY_METAFILE)


def _is_remote_deco(deco):
    return isinstance(deco, (BatchDecorator, KubernetesDecorator))


def _step_executes_remotely(step):
    "Check if a step is going to execute remotely or locally"
    return any(_is_remote_deco(deco) for deco in step.decorators)


def _is_env_deco(deco):
    "Check if a decorator is a known environment decorator (conda/pypi)"
    return isinstance(deco, (PyPIStepDecorator, CondaStepDecorator))


def _step_has_environment_deco(step):
    "Check if a step has a virtual environment decorator"
    return any(_is_env_deco(deco) for deco in step.decorators)



from metaflow.exception import MetaflowException


class SnowflakeException(MetaflowException):
    headline = "Snowflake error"


class SnowparkException(MetaflowException):
    headline = "Snowpark error"


class SnowparkKilledException(MetaflowException):
    headline = "Snowpark job killed"



import os
import sys
import platform

from metaflow import R, current
from metaflow.metadata import MetaDatum
from metaflow.metadata.util import sync_local_metadata_to_datastore
from metaflow.sidecar import Sidecar
from metaflow.decorators import StepDecorator
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    DEFAULT_CONTAINER_IMAGE,
    DEFAULT_CONTAINER_REGISTRY,
    SNOWPARK_ACCOUNT,
    SNOWPARK_USER,
    SNOWPARK_PASSWORD,
    SNOWPARK_ROLE,
    SNOWPARK_DATABASE,
    SNOWPARK_WAREHOUSE,
    SNOWPARK_SCHEMA,
)

from metaflow.metaflow_config import (
    DATASTORE_LOCAL_DIR,
)

from .snowpark_exceptions import SnowflakeException
from metaflow.plugins.aws.aws_utils import get_docker_registry


class Snowflake(object):
    def __init__(self, connection_params):
        self.connection_params = connection_params

    def session(self):
        # if using the pypi/conda decorator with @snowpark of any step,
        # make sure to pass {'snowflake': '0.11.0'} as well to that step
        try:
            from snowflake.snowpark import Session

            session = Session.builder.configs(self.connection_params).create()
            return session
        except (NameError, ImportError, ModuleNotFoundError):
            raise SnowflakeException(
                "Could not import module 'snowflake'.\n\nInstall Snowflake "
                "Python package (https://pypi.org/project/snowflake/) first.\n"
                "You can install the module by using the @pypi decorator - "
                "Eg: @pypi(packages={'snowflake': '0.11.0'})\n"
            )


class SnowparkDecorator(StepDecorator):
    name = "snowpark"

    defaults = {
        "account": None,
        "user": None,
        "password": None,
        "role": None,
        "database": None,
        "warehouse": None,
        "schema": None,
        "image": None,
        "stage": None,
        "compute_pool": None,
        "volume_mounts": None,
        "external_integration": None,
        "cpu": None,
        "gpu": None,
        "memory": None,
    }

    package_url = None
    package_sha = None
    run_time_limit = None

    def __init__(self, attributes=None, statically_defined=False):
        super(SnowparkDecorator, self).__init__(attributes, statically_defined)

        if not self.attributes["account"]:
            self.attributes["account"] = SNOWPARK_ACCOUNT
        if not self.attributes["user"]:
            self.attributes["user"] = SNOWPARK_USER
        if not self.attributes["password"]:
            self.attributes["password"] = SNOWPARK_PASSWORD
        if not self.attributes["role"]:
            self.attributes["role"] = SNOWPARK_ROLE
        if not self.attributes["database"]:
            self.attributes["database"] = SNOWPARK_DATABASE
        if not self.attributes["warehouse"]:
            self.attributes["warehouse"] = SNOWPARK_WAREHOUSE
        if not self.attributes["schema"]:
            self.attributes["schema"] = SNOWPARK_SCHEMA

        # If no docker image is explicitly specified, impute a default image.
        if not self.attributes["image"]:
            # If metaflow-config specifies a docker image, just use that.
            if DEFAULT_CONTAINER_IMAGE:
                self.attributes["image"] = DEFAULT_CONTAINER_IMAGE
            # If metaflow-config doesn't specify a docker image, assign a
            # default docker image.
            else:
                # Metaflow-R has its own default docker image (rocker family)
                if R.use_r():
                    self.attributes["image"] = R.container_image()
                # Default to vanilla Python image corresponding to major.minor
                # version of the Python interpreter launching the flow.
                self.attributes["image"] = "python:%s.%s" % (
                    platform.python_version_tuple()[0],
                    platform.python_version_tuple()[1],
                )

        # Assign docker registry URL for the image.
        if not get_docker_registry(self.attributes["image"]):
            if DEFAULT_CONTAINER_REGISTRY:
                self.attributes["image"] = "%s/%s" % (
                    DEFAULT_CONTAINER_REGISTRY.rstrip("/"),
                    self.attributes["image"],
                )

    # Refer https://github.com/Netflix/metaflow/blob/master/docs/lifecycle.png
    # to understand where these functions are invoked in the lifecycle of a
    # Metaflow flow.
    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # Set internal state.
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

        if any([deco.name == "parallel" for deco in decos]):
            raise MetaflowException(
                "Step *{step}* contains a @parallel decorator "
                "with the @snowpark decorator. @parallel is not supported with @snowpark.".format(
                    step=step
                )
            )

    def package_init(self, flow, step_name, environment):
        try:
            # Snowflake is a soft dependency.
            from snowflake.snowpark import Session
        except (NameError, ImportError, ModuleNotFoundError):
            raise SnowflakeException(
                "Could not import module 'snowflake'.\n\nInstall Snowflake "
                "Python package (https://pypi.org/project/snowflake/) first.\n"
                "You can install the module by executing - "
                "%s -m pip install snowflake\n"
                "or equivalent through your favorite Python package manager."
                % sys.executable
            )

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        if retry_count <= max_user_code_retries:
            cli_args.commands = ["snowpark", "step"]
            cli_args.command_args.append(self.package_sha)
            cli_args.command_args.append(self.package_url)
            cli_args.command_options.update(self.attributes)
            cli_args.command_options["run-time-limit"] = self.run_time_limit
            if not R.use_r():
                cli_args.entrypoint[0] = sys.executable

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        self.metadata = metadata
        self.task_datastore = task_datastore

        # this path will exist within snowpark container services
        login_token = open("/snowflake/session/token", "r").read()
        connection_params = {
            "account": os.environ.get("SNOWFLAKE_ACCOUNT"),
            "host": os.environ.get("SNOWFLAKE_HOST"),
            "authenticator": "oauth",
            "token": login_token,
            "database": os.environ.get("SNOWFLAKE_DATABASE"),
            "schema": os.environ.get("SNOWFLAKE_SCHEMA"),
            "autocommit": True,
            "client_session_keep_alive": True,
        }

        # SNOWFLAKE_WAREHOUSE is injected explicitly by us
        # but is not really required. So if it exists, we use it in
        # connection parameters
        if os.environ.get("SNOWFLAKE_WAREHOUSE"):
            connection_params["warehouse"] = os.environ.get("SNOWFLAKE_WAREHOUSE")

        snowflake_obj = Snowflake(connection_params)
        current._update_env({"snowflake": snowflake_obj})

        meta = {}
        if "METAFLOW_SNOWPARK_WORKLOAD" in os.environ:
            meta["snowflake-account"] = os.environ.get("SNOWFLAKE_ACCOUNT")
            meta["snowflake-database"] = os.environ.get("SNOWFLAKE_DATABASE")
            meta["snowflake-schema"] = os.environ.get("SNOWFLAKE_SCHEMA")
            meta["snowflake-host"] = os.environ.get("SNOWFLAKE_HOST")
            meta["snowflake-service-name"] = os.environ.get("SNOWFLAKE_SERVICE_NAME")

            self._save_logs_sidecar = Sidecar("save_logs_periodically")
            self._save_logs_sidecar.start()

        if len(meta) > 0:
            entries = [
                MetaDatum(
                    field=k,
                    value=v,
                    type=k,
                    tags=["attempt_id:{0}".format(retry_count)],
                )
                for k, v in meta.items()
                if v is not None
            ]
            # Register book-keeping metadata for debugging.
            metadata.register_metadata(run_id, step_name, task_id, entries)

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_retries
    ):
        if "METAFLOW_SNOWPARK_WORKLOAD" in os.environ:
            if hasattr(self, "metadata") and self.metadata.TYPE == "local":
                # Note that the datastore is *always* Amazon S3 (see
                # runtime_task_created function).
                sync_local_metadata_to_datastore(
                    DATASTORE_LOCAL_DIR, self.task_datastore
                )

        try:
            self._save_logs_sidecar.terminate()
        except:
            # Best effort kill
            pass

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]



import json
import os
from metaflow._vendor import click
from metaflow.cli import echo_always as echo
from metaflow.plugins.datastores.local_storage import LocalStorage

from .docker_environment import get_fastbakery_metafile_path
from .fast_bakery import FastBakeryApiResponse


@click.group()
def cli():
    pass


@cli.group(help="Commands related to Fast Bakery support.")
@click.pass_context
def fast_bakery(ctx):
    path = LocalStorage.get_datastore_root_from_config(echo, create_on_absent=False)
    ctx.obj.metafile_path = get_fastbakery_metafile_path(path, ctx.obj.flow.name)


@fast_bakery.command(help="Purge local Fast Bakery cache.")
@click.pass_obj
def purge(obj):
    try:
        os.remove(obj.metafile_path)
        echo("Local Fast Bakery cache purged.")
    except FileNotFoundError:
        echo("No local Fast Bakery cache found.")


@fast_bakery.command(help="List the cached images")
@click.pass_obj
def images(obj):
    current_cache = None
    try:
        with open(obj.metafile_path, "r") as f:
            current_cache = json.load(f)
    except FileNotFoundError:
        pass

    if current_cache:
        echo("List of locally cached image tags:\n")

        for val in current_cache.values():
            response = FastBakeryApiResponse(val)
            echo(response.container_image)

        echo(
            "In order to clear the cached images, you can use the command\n *fast-bakery purge*"
        )
    else:
        echo("No locally cached images.")
